@@ -11,7 +11,7 @@ from . import quadtree
 
 from .guppy import Point, Multipoint, Line, Polygon
 from .read import read_geojson, read_shapefile
-from ._shpfuncs import write_shapefile
+from .shp import write_shapefile
 from .gpx import GPX
 from .xyfile import load_xy, xyz2array_reg, array2xyz
 
