@@ -12,8 +12,8 @@ import numpy as np
 from . import vtk
 from . import geojson
 from . import xyfile
-from . import _shpfuncs
-from .. import crs as kcrs
+from . import shp
+from ..crs import crsreg, CRSError
 from .metadata import Metadata
 from . import _vectorgeo
 
@@ -23,18 +23,11 @@ except ImportError:
     sys.stderr.write("falling back on slow _vectorgeo")
     _vecgeo = _vectorgeo
 
-try:
-    import pyproj
-    PYPROJ = True
-    geod = pyproj.Geod(ellps="WGS84")
-except ImportError:
-    PYPROJ = False
-
 class Geometry(object):
     """ This is the abstract base class for all geometry types """
     _geotype = None
 
-    def __init__(self, crs=kcrs.CARTESIAN):
+    def __init__(self, crs=crsreg.CARTESIAN):
         self.properties = {}
         self._crs = crs
         return
@@ -42,14 +35,21 @@ class Geometry(object):
     def _distance(self, pos0, pos1):
         """ Generic method for calculating distance between positions that
         respects CRS """
-        if self._crs == kcrs.CARTESIAN:
+        if self._crs == crsreg.CARTESIAN:
             dist = _vecgeo.distance((pos0.x, pos0.y), (pos1.x, pos1.y))
-        elif PYPROJ:
-            _, _, dist = geod.inv(pos0.x, pos0.y, pos1.x, pos1.y, radians=False)
+        elif pos0._crs == pos1._crs:
+            try:
+                _, _, dist = pos0._crs.geod.inv(pos0.x, pos0.y, pos1.x, pos1.y, radians=False)
+            except NameError:
+                dist = greatcircle(pos0, pos1)
         else:
-            dist = greatcircle(pos0, pos1)
+            raise CRSError("Positions must use the same CRS")
         return dist
 
+    @property
+    def crs(self):
+        return self._crs
+
     def add_property(self, name, value):
         """ Insert a property (name -> value) into the properties dict, raising
         a NameError if the property already exists. Compared to directly
@@ -153,13 +153,7 @@ class Point(Geometry):
         if self.coordsxy() == other.coordsxy():
             return np.nan
 
-        if self._crs == kcrs.LONLAT and other._crs == kcrs.LONLAT:
-            if not PYPROJ:
-                raise CRSError("Non-cartesian points require pyproj")
-            az1, _, _ = geod.inv(self.x, self.y, other.x, other.y)
-            return az1 * math.pi / 180.0
-
-        elif self._crs == kcrs.CARTESIAN and other._crs == kcrs.CARTESIAN:
+        if self._crs == self._crs == crsreg.CARTESIAN:
             dx = other.x - self.x
             dy = other.y - self.y
 
@@ -183,6 +177,10 @@ class Point(Geometry):
                 else:
                     return math.pi
 
+        elif self._crs == other._crs:
+            az1, _, _ = self._crs.geod.inv(self.x, self.y, other.x, other.y)
+            return az1 * math.pi / 180.0
+
         else:
             raise CRSError("Azimuth undefined for points in CRS {0} and "
                            "{1}".format(self._crs, other._crs))
@@ -194,7 +192,7 @@ class Point(Geometry):
             distance (float): distance to walk
             direction (float): horizontal walk direction in radians
         """
-        if self._crs == kcrs.CARTESIAN:
+        if self._crs == crsreg.CARTESIAN:
             dx = distance * math.sin(direction)
             dy = distance * math.cos(direction)
 
@@ -205,15 +203,13 @@ class Point(Geometry):
                 return Point((self.x+dx, self.y+dy, self.z),
                              properties=self.properties, data=self.data, crs=self._crs)
 
-        elif PYPROJ:
-            (x, y, backaz) = geod.fwd(self.x, self.y, direction*180.0/math.pi,
-                                      distance, radians=False)
+        else:
+            (x, y, backaz) = self._crs.geod.fwd(self.x, self.y,
+                                                direction*180.0/math.pi,
+                                                distance, radians=False)
             return Point((x, y), properties=self.properties, data=self.data,
                          crs=self._crs)
 
-        else:
-            raise CRSError("Non-cartesian points require pyproj")
-
     def distance(self, other):
         """ Returns a distance to another Point. If the coordinate system is
         geographical and a third (z) coordinate exists, it is assumed to have
@@ -228,12 +224,10 @@ class Point(Geometry):
 
     def greatcircle(self, other):
         """ Return the great circle distance between two geographical points. """
-        if not PYPROJ:
-            raise CRSError("Non-cartesian points require pyproj")
-        if not (self._crs == kcrs.LONLAT and other._crs == kcrs.LONLAT):
-            raise CRSError("Great circle distances require both points to be "
-                           "in geographical coordinates")
-        az1, az2, dist = geod.inv(self.x, self.y, other.x, other.y, radians=False)
+        if not (self._crs == other._crs):
+            raise CRSError("Both point have use the same CRS")
+        az1, az2, dist = self._crs.geod.inv(self.x, self.y, other.x, other.y,
+                                            radians=False)
         return dist
 
     def shift(self, shift_vector):
@@ -293,17 +287,31 @@ class MultipointBase(Geometry):
 
     def __init__(self, vertices, data=None, properties=None, copy_metadata=True,
                  **kwargs):
+        """ Partial init function that establishes geometry rank and greates a
+        metadata attribute.
+        """
         super(MultipointBase, self).__init__(**kwargs)
         vertices = list(vertices)
         if len(vertices) > 0:
-            self.rank = len(vertices[0])
+
+            def getrank(vertex):
+                try:
+                    return vertex.rank
+                except AttributeError:
+                    return len(vertex)
+
+            self.rank = getrank(vertices[0])
+            try:
+                self._crs = vertices[0].crs
+            except AttributeError:
+                pass
 
             if not 2 <= self.rank <= 3:
                 raise GInitError("Input must be doubles or triples")
-            elif not all(self.rank == len(i) for i in vertices):
+            elif not all(self.rank == getrank(v) for v in vertices):
                 raise GInitError("Input must have consistent rank")
             else:
-                self.vertices = [tuple(i) for i in vertices]
+                self.vertices = [tuple(v) for v in vertices]
 
         else:
             self.rank = None
@@ -453,6 +461,15 @@ class MultipointBase(Geometry):
         self.shift(origin)
         return self
 
+    def apply_affine_transform(self, M):
+        """ Apply an affine transform given by matrix *M* to data and return a
+        new geometry. """
+        vertices = []
+        for x,y in self.get_vertices():
+            vertices.append(tuple(np.dot(M, [x, y, 1])[:2]))
+        return type(self)(vertices, data=self.data, properties=self.properties,
+                          crs=self._crs)
+
     def _subset(self, idxs):
         """ Return a subset defined by index in *idxs*. """
         vertices = [self.vertices[i] for i in idxs]
@@ -584,9 +601,9 @@ class MultipointBase(Geometry):
     def to_shapefile(self, fstem):
         """ Save line to a shapefile """
         if self.rank == 2:
-            _shpfuncs.write_multipoint2(self, fstem)
+            shp.write_multipoint2(self, fstem)
         elif self.rank == 3:
-            _shpfuncs.write_multipoint3(self, fstem)
+            shp.write_multipoint3(self, fstem)
         else:
             raise IOError("Rank must be 2 or 3 to write as a shapefile")
         return
@@ -608,6 +625,9 @@ class Multipoint(MultipointBase):
                  **kwargs):
         def ispoint(a):
             return hasattr(a, "_geotype") and a._geotype == "Point"
+
+        vertices = list(vertices)
+
         if len(vertices) != 0 and all(ispoint(a) for a in vertices):
             points = vertices
             crs = points[0]._crs
@@ -638,7 +658,6 @@ class Multipoint(MultipointBase):
             self._crs = crs
 
         else:
-
             super(Multipoint, self).__init__(vertices,
                                              data=data,
                                              properties=properties,
@@ -668,16 +687,19 @@ class Multipoint(MultipointBase):
 class ConnectedMultipoint(MultipointBase):
     """ Class for Multipoints in which vertices are assumed to be connected. """
 
+    @property
     def length(self):
         """ Returns the length of the line/boundary. """
         points = [Point(i, crs=self._crs) for i in self.vertices]
         distances = [a.distance(b) for a, b in zip(points[:-1], points[1:])]
         return sum(distances)
 
+    @property
     def segments(self):
         """ Returns an generator of adjacent line segments. """
         return (self._subset((i,i+1)) for i in range(len(self)-1))
 
+    @property
     def segment_tuples(self):
         """ Returns an generator of adjacent line segments as coordinate tuples. """
         return ((self.vertices[i], self.vertices[i+1])
@@ -687,7 +709,7 @@ class ConnectedMultipoint(MultipointBase):
         """ Return whether an intersection exists with another geometry. """
         interxbool = (np.nan in _vecgeo.intersection(a[0][0], a[1][0], b[0][0], b[1][0],
                                                      a[0][1], a[1][1], b[0][1], b[1][1])
-                    for a in self.segments() for b in other.segments())
+                    for a in self.segments for b in other.segments)
         if self._bbox_overlap(other) and (True not in interxbool):
             return True
         else:
@@ -697,7 +719,7 @@ class ConnectedMultipoint(MultipointBase):
         """ Return the intersections with another geometry as a Multipoint. """
         interx = (_vecgeo.intersection(a[0][0], a[1][0], b[0][0], b[1][0],
                                           a[0][1], a[1][1], b[0][1], b[1][1])
-                     for a in self.segments() for b in other.segments())
+                     for a in self.segments for b in other.segments)
         if not keep_duplicates:
             interx = set(interx)
         interx_points = []
@@ -716,12 +738,13 @@ class ConnectedMultipoint(MultipointBase):
         """
         if self._crs != pt._crs:
             raise CRSError("Point must have matching CRS")
-        elif self._crs == kcrs.CARTESIAN:
+        elif self._crs == crsreg.CARTESIAN:
             func = _vecgeo.pt_nearest_planar
         else:
+            geod = self._crs.geod
             func = lambda *args: _vecgeo.pt_nearest_proj(geod, *args, tol=0.01)
 
-        segments = list(self.segments())
+        segments = list(self.segments)
         point_dist = map(func,
                          (pt.vertex for _ in segments),
                          (seg[0].vertex for seg in segments),
@@ -735,23 +758,24 @@ class ConnectedMultipoint(MultipointBase):
         """
         if self._crs != pt._crs:
             raise CRSError("Point must have matching CRS")
-        elif self._crs == kcrs.CARTESIAN:
+        elif self._crs == crsreg.CARTESIAN:
             func = _vecgeo.pt_nearest_planar
         else:
+            geod = self._crs.geod
             func = lambda *args: _vecgeo.pt_nearest_proj(geod, *args, tol=0.01)
 
-        segments = list(self.segments())
+        segments = list(self.segments)
         point_dist = list(map(func,
                               (pt.vertex for _ in segments),
                               (seg[0].vertex for seg in segments),
                               (seg[1].vertex for seg in segments)))
         distances = [i[1] for i in point_dist]
-        return Point(point_dist[distances.index(min(distances))][0],
-                     properties=self.properties, crs=self._crs)
+        imin = distances.index(min(distances))
+        return Point(point_dist[imin][0], crs=self._crs)
 
     def within_distance(self, pt, distance):
         """ Test whether a point is within *distance* of a ConnectedMultipoint. """
-        return all(distance >= seg.shortest_distance_to(pt) for seg in self.segments())
+        return all(distance >= seg.shortest_distance_to(pt) for seg in self.segments)
 
 
 class Line(ConnectedMultipoint):
@@ -815,7 +839,7 @@ class Line(ConnectedMultipoint):
 
     def subsection(self, n):
         """ Return *n* equally spaced Point instances along line. """
-        segments = self.segments()
+        segments = self.segments
         Ltotal = self.cumlength()[-1]
         step = Ltotal / float(n-1)
         step_remaining = step
@@ -885,9 +909,9 @@ class Line(ConnectedMultipoint):
     def to_shapefile(self, fstem):
         """ Save line to a shapefile """
         if self.rank == 2:
-            _shpfuncs.write_line2(self, fstem)
+            shp.write_line2(self, fstem)
         elif self.rank == 3:
-            _shpfuncs.write_line3(self, fstem)
+            shp.write_line3(self, fstem)
         else:
             raise IOError("rank must be 2 or 3 to write as a shapefile")
         return
@@ -907,7 +931,9 @@ class Polygon(ConnectedMultipoint):
     subs = []
 
     def __init__(self, vertices, data=None, properties=None, subs=None, **kwargs):
-        ConnectedMultipoint.__init__(self, vertices, data=data, properties=properties, **kwargs)
+        vertices = list(vertices)
+        ConnectedMultipoint.__init__(self, vertices, data=data,
+                                     properties=properties, **kwargs)
         if vertices[0] != vertices[-1]:
             self.vertices.append(vertices[0])
         self.subs = subs if subs is not None else []
@@ -933,11 +959,19 @@ class Polygon(ConnectedMultipoint):
                       crs=self._crs, copy_metadata=False)
         return subset
 
+    def isclockwise(self):
+        """ Return whether polygon winds clockwise around its interior. """
+        s = sum((seg[1][0] - seg[0][0]) * (seg[1][1] + seg[0][1])
+                for seg in self.segments)
+        return s > 0
+
+    @property
     def perimeter(self):
         """ Return the perimeter of the polygon. If there are sub-polygons,
         their perimeters are added recursively. """
-        return self.length() + sum([p.perimeter() for p in self.subs])
+        return self.length + sum([p.perimeter for p in self.subs])
 
+    @property
     def area(self):
         """ Return the two-dimensional area of the polygon. If there are
         sub-polygons, their areas are subtracted. """
@@ -945,7 +979,7 @@ class Polygon(ConnectedMultipoint):
         for i in range(len(self.vertices)-1):
             a += 0.5 * abs((self.vertices[i][0] + self.vertices[i+1][0])
                          * (self.vertices[i][1] - self.vertices[i+1][1]))
-        return a - sum(map(lambda p: p.area(), self.subs))
+        return a - sum(map(lambda p: p.area, self.subs))
 
     def contains(self, pt):
         """ Returns True if pt is inside or on the boundary of the
@@ -953,7 +987,7 @@ class Polygon(ConnectedMultipoint):
         """
         nintx = 0
         x, y = pt[0], pt[1]
-        for seg in self.segments():
+        for seg in self.segments:
             (a, b) = seg
             if _vecgeo.intersects_cn(x, y, a[0], b[0], a[1], b[1]):
                 nintx += 1
@@ -968,9 +1002,9 @@ class Polygon(ConnectedMultipoint):
     def to_shapefile(self, fstem):
         """ Save line to a shapefile """
         if self.rank == 2:
-            _shpfuncs.write_poly2(self, fstem)
+            shp.write_poly2(self, fstem)
         elif self.rank == 3:
-            _shpfuncs.write_poly3(self, fstem)
+            shp.write_poly3(self, fstem)
         else:
             raise IOError("rank must be 2 or 3 to write as a shapefile")
         return
@@ -1002,12 +1036,6 @@ class GGeoError(GuppyError):
         self.message = message
 
 
-class CRSError(GuppyError):
-    """ Exception to raise for invalid geodetic operations. """
-    def __init__(self, message=''):
-        self.message = message
-
-
 def ray_intersection(pt, endpt1, endpt2, direction=0.0):
     """ Determines whether a ray intersects a line segment. If yes,
     returns the point of intersection. If no, return None. Input
@@ -1096,15 +1124,6 @@ def greatcircle(pta, ptb, method="vicenty"):
     return distance
 
 
-def sortby(A, B):
-    """ Sort a list A by the values in an ordered list B. """
-    if len(A) != len(B):
-        raise GGeoError("A and B must be of the same length")
-    comb = zip(B,A)
-    comb.sort()
-    return [i[1] for i in comb]
-
-
 def points_to_multipoint(points):
     """ Merge *points* into a Multipoint instance. Point properties are stored
     as Multipoint data. All points must use the same CRS.
@@ -1127,3 +1146,15 @@ def points_to_multipoint(points):
 
     return Multipoint(vertices, data=ptdata, crs=crs)
 
+def affine_matrix(mpa, mpb):
+    """ Compute the affine transformation matrix that projects Multipoint mpa
+    to Multipoint mpb using a least squares fit. """
+    if len(mpa) != len(mpb):
+        raise GuppyError("Input geometries must have identical length")
+    vecp = np.asarray(mpb.get_vertices()).ravel()
+    A = np.empty([2*len(mpa), 6], dtype=np.float64)
+    for i, (x, y) in enumerate(mpa.get_vertices()):
+        A[2*i:2*i+2,:] = np.kron(np.eye(2), [x, y, 1])
+    M, res, rank, singvals = np.linalg.lstsq(A, vecp)
+    return np.vstack([np.reshape(M, [2, 3]), np.atleast_2d([0, 0, 1])])
+
